#include <stdio.h>
#include <stdlib.h>
#include "inc/opucem.h"

// **************************************************
// Formulas from Forshaw, Ross, White, hep-ph/0107232
// **************************************************

double THSM ( double Z, double s2w, double h) {
  return 3./(16*M_PI*(1-s2w)*s2w) * ( h*h/(Z*Z-h*h)*2*log(h/Z) - (1-s2w)*h*h/((1-s2w)*Z*Z-h*h)*log(h*h/(1-s2w)/Z/Z) );
}

double SHSM ( double Z, double h) {
  double multiplier=0.;
  if ( h < 2*Z) {
    multiplier=sqrt(4*Z*Z-h*h)*atan(sqrt((4*Z*Z-h*h)/h/h));
  } else {
    multiplier=sqrt(h*h-4*Z*Z)*log(2*Z/(h+sqrt(h*h-4*Z*Z)));
  }

  return 1./M_PI* (   3/8.*(h*h/Z/Z) - 1/12.*pow(h/Z,4)
		    + pow(h/Z,2)*2*log(h/Z)*((3*Z*Z-h*h)/4/Z/Z+1./24.*pow(h/Z,4)+3./4.*Z*Z/(Z*Z-h*h))
		    + (1-1/3.*h*h/Z/Z+1/12.*pow(h/Z,4))*h/Z/Z*multiplier );
}


// **********************************************
// Formulas from He, Polonsky, Su, hep-ph/0102144
// **********************************************

double F ( double x1, double x2 ) {
  // We added the equality case, using l'Hpital's rule
  if ( x1 == x2 ) return 0;
  return ( (x1+x2)/2.0 - log(x1/x2)*(x1*x2)/(x1-x2));
}

double G ( double x ) {
  return ( -4*sqrt(4*x -1)*atan(1/sqrt(4*x-1)));
}

double f ( double x1, double x2 ) {
  double Delta = 2*(x1+x2) - (x1-x2)*(x1-x2) -1;

  if (Delta == 0) return 0;
  if (Delta  < 0)
    return ( sqrt(-Delta)*log((x1+x2-1+sqrt(-Delta))/(x1+x2-1-sqrt(-Delta)) ));
  //if (Delta  > 0)
  return ( -2*sqrt(Delta)*(atan((x1-x2+1)/sqrt(Delta)) - atan((x1-x2-1)/sqrt(Delta)) ) );
}

double bet22 ( double qsq, double m1sq, double m2sq ) {
  double x1=m1sq/qsq;
  double x2=m2sq/qsq;

  if (m1sq == m2sq) {
    return ( (qsq/24)*
	     (2*log(qsq)+2*log(x1) + (16*x1 - 10./3.) + (4*x1-1)*G(x1))
	     );
  }
  return ( (qsq/24)*
	   (2*log(qsq)+log(x1*x2)
	    +( pow(x1-x2,3) -3*(x1*x1 -x2*x2) +3*(x1-x2))*log(x1/x2)
	    -( 2*(x1-x2)*(x1-x2) -8*(x1+x2) +10./3.)
	    -( (x1-x2)*(x1-x2) -2*(x1+x2) +1)*f(x1,x2) -6*F(x1,x2) ) 
	   );
}

double bet0( double qsq, double m1sq, double m2sq ) {
  double x1=m1sq/qsq;
  double x2=m2sq/qsq;
  if (m1sq == m2sq) {
    return ( 2-2*sqrt(4*x1-1)*atan(1/sqrt(4*x1-1)));
  }
  return ( 1+0.5*( (x1+x2)/(x1-x2) -(x1-x2))*log(x1/x2) +0.5*f(x1,x2)); 
}


double bbar0( double m1sq, double m2sq, double m3sq ) {
  return (
	  (m1sq*log(m1sq) -m3sq*log(m3sq))/(m1sq-m3sq) 
	  -(m1sq*log(m1sq) -m2sq*log(m2sq))/(m1sq-m2sq) 
	  );
}

double SH_2HDM ( double Z,
		 double H, double A, double Hc, double h,
		 double b, double a ) {
  return 1/(M_PI*Z*Z)*(
		       sin(b-a)*sin(b-a)*bet22(Z*Z, H*H, A*A) -bet22(Z*Z, Hc*Hc, Hc*Hc)
		       +cos(b-a)*cos(b-a)*( bet22(Z*Z, h*h, A*A) +bet22(Z*Z, Z*Z, H*H) -bet22(Z*Z, Z*Z, h*h)
					    -Z*Z*bet0(Z*Z,Z*Z, H*H) +Z*Z*bet0(Z*Z,Z*Z,h*h)) 
		       );
}

// Our modified SH. Removed all terms which have H,A,Hc. Multiplied by -1. Took cos(b-a)={1,-1}
// Gives the same results as SHSM()
double SH ( double Z, double h ) {
  return -1./(M_PI*Z*Z) * ( -bet22(Z*Z, Z*Z, h*h) +Z*Z*bet0(Z*Z,Z*Z,h*h) );
}

double TH_2HDM ( double Z, double W, double s2w,
		 double H, double A, double Hc, double h,
		 double b, double a ) {

  return 1/(16*M_PI*W*W*s2w)*(
			      F(Hc*Hc, A*A) + sin(b-a)*sin(b-a)*( F(Hc*Hc, H*H) - F(A*A, H*H) )
			      + cos(b-a)*cos(b-a)*( F(Hc*Hc, h*h) - F(A*A, H*H) - F(W*W, h*h) 
						    - F(Z*Z, H*H) + F(Z*Z, h*h) + 4*Z*Z*bbar0(Z*Z, H*H, h*h) - 4*W*W*bbar0(W*W, H*H, h*h) )
			      );
}

// Our modified TH. Removed all terms which have H,A,Hc. Multiplied by 3. Took cos(b-a)={1,-1}
// It gives a difference of about -0.025 when we move from mH=114 to mH=1000, in agreement with the LEPEWWG plot
//   from http://lepewwg.web.cern.ch/LEPEWWG/plots/summer2006/s06_stu_contours.eps
// It is also in agreement with fig.1 of He, et.al.
// Gives the same results as THSM()
double TH ( double Z, double W, double s2w, double h) {
  return 3./(16*M_PI*W*W*s2w)*( - F(W*W, h*h) + F(Z*Z, h*h) ) ;
}

// IMPORTANT! He et.al. define hypercharge using Q=I_z+Y. For this reason Y=1/6 (-1/2) for quarks (leptons).
double SF ( double Y, double Nc, double m1, double m2, double Z ) {
  double x1=(m1/Z)*(m1/Z);
  double x2=(m2/Z)*(m2/Z);
  return Nc/(6*M_PI)*( 2*(4*Y+3)*x1 + 2*(-4*Y+3)*x2 - 2*Y*log(x1/x2)
		       + ((1.5+2*Y)*x1+Y)*G(x1) + ((1.5-2*Y)*x2-Y)*G(x2) );
}

double TF ( double Nc, double m1, double m2, double Z, double s2w ) {
  double x1=(m1/Z)*(m1/Z);
  double x2=(m2/Z)*(m2/Z);
  return Nc/(8.*M_PI*s2w*(1-s2w)) * F(x1,x2);
}

// This formula was checked against fig 2(c).
double UF ( double Nc, double m1, double m2, double Z ) {
  if ( m1 == m2 ) return 0;  // Limiting case added manually. Oct 8, 2009, veo.
  double x1=(m1/Z)*(m1/Z);
  double x2=(m2/Z)*(m2/Z);
  return Nc/(-2*M_PI)*(   (x1+x2)/2 - (x1-x2)*(x1-x2)/3 + (pow(x1-x2,3)/6-0.5*(x1*x1+x2*x2)/(x1-x2))*log(x1/x2)
			+ (x1-1)/6*f(x1,x1) + (x2-1)/6*f(x2,x2) + (1/3.-(x1+x2)/6-(x1-x2)*(x1-x2)/6)*f(x1,x2) );
}


// ***********************************************************
// Majorana related formulae from Gates & Terning, PRL67, 1991
// ***********************************************************

double f1 ( double m1, double m2 ) {
  return (3*m1*m2*m2*m2 + 3*m1*m1*m1*m2 - 4*m1*m1*m2*m2) / pow((m1*m1-m2*m2),2); }

double f2 ( double m1, double m2 ) {
  return ( pow(m1,6) - 3*pow(m1,4)*m2*m2 + 6*pow(m1*m2,3) - 3*m1*m1*pow(m2,4) + pow(m2,6) ) / pow(m1*m1-m2*m2,3) ; }

double SM ( double m1, double m2, double mE ) {
  double ctsq=pow(m2/(m1+m2),1);  // Power reduced to 1: Gates and Terning has an incorrect equation 10!
  double stsq=pow(m1/(m1+m2),1);
  return 1/(6*M_PI)*( ctsq*2*log(m1/mE) + stsq*2*log(m2/mE) + 1.5 - stsq*ctsq*(8./3.+ f1(m1,m2) - f2(m1,m2)*2*log(m1/m2)) ); }

double TM ( double m1, double m2, double mE, double s2w, double W ) {
  double ctsq=pow(m2/(m1+m2),1);  // Power reduced to 1: Gates and Terning has an incorrect equation 10!
  double stsq=pow(m1/(m1+m2),1);
  return 1/(16*M_PI*s2w*W*W)*(   ctsq*(m1*m1+mE*mE-(2*m1*m1*mE*mE)/(m1*m1-mE*mE)*2*log(m1/mE))
			       + stsq*(m2*m2+mE*mE-(2*m2*m2*mE*mE)/(m2*m2-mE*mE)*2*log(m2/mE))
			       - stsq*ctsq*(m1*m1+m2*m2-4*m1*m2 + 2*(m1*m1*m1*m2-m1*m1*m2*m2+m1*m2*m2*m2)*2*log(m1/m2)/(m1*m1-m2*m2)) ) ;
}


// *******************************************************************
// Majorana related approx and exact formulae from Kniehl & Kohrs, PRD
// *******************************************************************

double A( double m1, double m2 ) {
  // Limiting case of m1==m2 computed with Mathematica manually (Oct 5, 2009, veo)
  if ( m1 == m2 )  return 2*m1*m1/3;
  return m1*m1*m2*m2/pow(m1*m1-m2*m2,2)*(-3*(m1*m1+m2*m2)+4*m1*m2)
    + m1*m2/(m1*m1-m2*m2)*(m1*m1+m2*m2+2*m1*m1*m2*m2/pow(m1*m1-m2*m2,2)*(3*m1*m2-m1*m1-m2*m2))*2*log(m1/m2) ;
}

// SMapprox() gives the same result as SM() a la Gates & Terning, but we use the modified A(), so it also works when m1==m2.
// It gives 0.026807 as the result to inputs 100, 200, 300, in agreement with fig.1
double SMapprox( double m1, double m2, double mE ) {
  return 1/(6*M_PI)/(m1+m2)*(
			       1/(m1+m2)*(1.5*(m1*m1+m2*m2)+(m1*m2/3)+A(m1,m2))
			     + m2*2*log(m1/mE) + m1*2*log(m2/mE) );
}

// UMapprox gives -0.00180571 as the result of 100,200,300, in agreement with fig.1
double UMapprox( double m1, double m2, double mE ) {
  return 1/(6*M_PI)/(m1+m2)*(
			     -1/(m1+m2)*(13*(m1*m1+m2*m2)/6+5*m1*m2/3+A(m1,m2))
			     +m2/pow(m1*m1-mE*mE,2)*(4*pow(m1*mE,2)+(m1*m1+mE*mE)*(pow(m1,4)+pow(mE,4)-4*pow(m1*mE,2))/(m1*m1-mE*mE)*2*log(m1/mE)) 
			     +m1/pow(m2*m2-mE*mE,2)*(4*pow(m2*mE,2)+(m2*m2+mE*mE)*(pow(m2,4)+pow(mE,4)-4*pow(m2*mE,2))/(m2*m2-mE*mE)*2*log(m2/mE)) );
}

// TMexact() gives the same result as TM() a la Gates & Terning
double TMexact( double m1, double m2, double mE, double s2w, double W ) {
  // Limiting equality cases computed with Mathematica manually (Nov 2, 2009, veo)
  double m1mEterm = m1==mE ? -2*mE : (m2-m1*(m1*m1+mE*mE)/(m1*m1-mE*mE))*2*log(m1/mE);
  double m2mEterm = m2==mE ? -2*mE : (m1-m2*(m2*m2+mE*mE)/(m2*m2-mE*mE))*2*log(m2/mE);
  double m1m2term = m1==m2 ? 4*m1*m1 : 6*m1*m2+((pow(m1,4)+pow(m2,4)-2*m1*m2*(m1*m1+m2*m2))/(m1*m1-m2*m2))*2*log(m1/m2);
  return 1/(16*M_PI*s2w*W*W)*(   mE*mE
			       + m1*m2/pow(m1+m2,2) * m1m2term
			       + m1*m2/(m1+m2) * ( m1mEterm + m2mEterm ) );
}

double complex lambda( double complex x, double y, double z) {
  return x*x + y*y + z*z - 2*(x*y+y*z+z*x);
}

long double complex B0(double s1, double m1sq, double m2sq) {
  const double C = 0.5772156649; // Euler-Mascheroni constant
  long double Delta = 2/little/little-C-log(M_PI); // 2/(4-n)-C-log(M_PI);

  if (s1==0 && m1sq==m2sq) return Delta - logl(m1sq);

  // Imaginary part of lambda(s1+I*epsilon,m1sq,m2sq) changes sign when m1sq+m2sq<s1. This causes a sign change in the
  //  imaginary part of csqrtl(lambda(...)) term, which ends up changing the real part of B0(...) itself.
  double epssign = (m1sq+m2sq<s1) ? -1 : 1;   // Accounts for lambda(...) sign issue.  veo, oct 11, 2009.
  return   Delta - 0.5*logl(m1sq*m2sq) + 2
         + 1/s1*(  -0.5*(m1sq-m2sq)*logl(m1sq/m2sq)
		 + csqrtl(lambda(s1+I*epsilon*epssign,m1sq,m2sq)) * cacoshl((m1sq+m2sq-s1-I*epsilon)/(2*sqrtl(m1sq*m2sq))) );
}

long double complex PI( int vasign, double qsq, double m1, double m2 ) {
  if ( abs(vasign)!=1 ) {
    printf("PI function called with wrong vasign value!\n");
    exit(1); }
  // the first term modified to qsq from s, a la Natale-DaSilva MPLA'95. KK paper is incorrect.
  return 1/(12*M_PI*M_PI)*(
			   ( qsq - (m1*m1+m2*m2)/2 + vasign*3*m1*m2 - pow(m1*m1-m2*m2,2)/(2*qsq) ) * B0(qsq,m1*m1,m2*m2)
			   + m1*m1*(-1+0.5*(m1*m1-m2*m2)/qsq) * B0(0,m1*m1,m1*m1)
			   + m2*m2*(-1+0.5*(m2*m2-m1*m1)/qsq) * B0(0,m2*m2,m2*m2)
			   - qsq/3 + 0.5*pow(m1*m1-m2*m2,2)/qsq ) ;
}

long double complex PI_V( double qsq, double m1, double m2 )  { return PI(  1, qsq, m1, m2 ); }
long double complex PI_A( double qsq, double m1, double m2 )  { return PI( -1, qsq, m1, m2 ); }

// Highly recommend using the wrapper SMexactWithChecks() instead of SMexact() itself. See below!
// SMexact() can exactly reproduce fig1.
// For ksi=1, it gives : 0.021802, and as ksi increases, it converges on to SMapprox().
double SMexact( double m1, double m2, double mE, double s2w, double z ) {
  long double ctsq=m2/(m1+m2);
  long double stsq=m1/(m1+m2);

  return M_PI/z*(
		 -2*ctsq*ctsq*( creall( PI_A(z,m1,m1) ) - PI_A(0+little,m1,m1) )   // creall's are actually not needed
		 -2*stsq*stsq*( creall( PI_A(z,m2,m2) ) - PI_A(0+little,m2,m2) )   // since typecasting to double takes
		 -4*ctsq*stsq*( creall( PI_V(z,m1,m2) ) - PI_V(0+little,m1,m2) )   // only the real part of complex nos

		 //+3* creall( PI_V(z,mE,mE) )    // the version that appears in the paper
		 +3* ( creall( PI_V(z,mE,mE) ) - PI_V(0+little,mE,mE) )  // adding PI_V(0,...) term improves the result
		 // sin2thetaW runs between q2=mZ2 down to q2=0. Take the Kappa from Ferroglia et.al., hep-ph/0307200v2 :
		 //-pow(4*s2w-1,2)* creall( PI_V(z,mE,mE) ) + pow(4*s2w*1.0317/1.0028-1,2)* PI_V(0+little,mE,mE)
		 -1* ( creall( PI_A(z,mE,mE) ) - PI_A(0+little,mE,mE) )
		 );
}


// S+U exact formula, checked against fig1: SpUMexact()-SMexact() replicates fig1 successfully.
// At ksi=1, SpUMexact()-SMexact() gives 0.00623479, and it slowly converges onto UMapprox() as in fig1.
// At ksi=5, SpUMexact()-SMexact() gives -0.00163245, in good agreement with the UMapprox() value of -0.00180571
// SpUMexact()-SMexact() is also checked against SF() when m1=m2 and seems ok for lowish masses - at higher masses UMapprox() looks better
double SpUMexact( double m1, double m2, double mE, double w, double z ) {
  long double ctsq=m2/(m1+m2);
  long double stsq=m1/(m1+m2);
  return 2*M_PI*(
		 -1*ctsq/w* ( creall( PI_V(w,m1,mE)+PI_A(w,m1,mE) ) - (PI_V(0+little,m1,mE)+PI_A(0+little,m1,mE)) )
		 -1*stsq/w* ( creall( PI_V(w,m2,mE)+PI_A(w,m2,mE) ) - (PI_V(0+little,m2,mE)+PI_A(0+little,m2,mE)) )
		 //+2/z * ( creall( PI_V(z,mE,mE) ) )
		 +2/z * ( creall( PI_V(z,mE,mE) ) - PI_V(0+little,mE,mE) )  // adding PI_V(little,...) slightly improves agreement with UF()
		 );
}


// **************************************
// Test methods for the Majorana formulae
// **************************************

long double testB0( double m ) {
  if ( epsilon > little / 100 )   // I could also check that imag(B0(...)) << real(B0(...))
    printf("Your B0 implementation might have a large epsilon (%Lg) compared to q^2 scale of %Lg\n", epsilon, little );
  long double test = creall(B0(0, m*m, m*m));
  if ( test < 0 )
    printf("Your B0 implementation might have a very small Delta at mass scale %g (B0=%Lg)\n", m, test );
  test = test / creall(B0(0+little, m*m, m*m)) - 1;
  if ( fabs(test) > 5e-6 )
    printf("Your B0 implementation is not reliable at mass scale = %g (error~=%Lg)\n", m, test );
  return test; }

// SM exact should give the same results as SF, when m1==m2, equivalent of Dirac case
void testSMvsSF( double m, double mE, double s2w, double Z ){
  double sf = SF(-0.5, 1, m, mE, Z);
  double sm = SMexact(m, m, mE, s2w, Z*Z);
  if ( fabs(sf/sm-1) > 1e-4  ||  fabs(sf-sm) > 1e-4 )
    printf("Testing failed for m1=m2=%g, mE=%g, Sdirac=%g\tSmaj=%g\n", m, mE, sf, sm );
}

double SMexactWithChecks( double m1, double m2, double mE, double s2w, double Z ){

  // Before we move on to the actual calculation, let's test if our epsilon and little are OK.
  testB0(m1); testB0(m2);

  // for the two neutrinos masses check that maj->dirac
  testSMvsSF( m1, mE, s2w, Z );
  testSMvsSF( m2, mE, s2w, Z );

  // Now test against SM approx.
  double smapprox = SMapprox(m1, m2, mE);
  double smexact  = SMexact(m1, m2, mE, s2w, Z*Z);
  if ( fabs(smapprox/smexact-1) > 0.25  ||  fabs(smapprox-smexact) > 0.1 )
    printf("Warning! Approx and exact formulas differ significantly! Sa=%g\tSe=%g\n", smapprox, smexact );
  return smexact;
}

double UMexactWithChecks( double m1, double m2, double mE, double s2w, double Z, double smexact ) {
  double w = Z*Z*(1-s2w);
  // if smexact has been provided for us, we use it, otherwise we compute:
  if ( smexact == 0 )
    smexact = SMexactWithChecks( m1, m2, mE, s2w, Z );

  // Test against UM approx.
  double umapprox = UMapprox(m1,m2,mE);
  double umexact  = SpUMexact(m1,m2,mE,w,Z*Z) - smexact;
  if ( fabs(umapprox-umexact) > 0.05 )
    printf("Warning! Approx and exact formulas differ significantly! Ua=%g\tUe=%g\n", umapprox, umexact );
  return umexact;
}


// ******************************
// Other functions of convenience
// ******************************

// Contributions from the changes in the reference Higgs and top-quark masses
double SshiftFromRef( double h, double href, double t, double tref, double Z ) {
  double shift = SF(1/6., 3, t, 500, Z) - SF(1/6., 3, tref, 500, Z);  // Sshift is independent of b-quark mass
  shift += SH(Z,h) - SH(Z,href);
  return shift;
}

double TshiftFromRef( double h, double href, double t, double tref, double Z, double s2w ) {
  double shift = TF(3, t, 4.5, Z, s2w) - TF(3, tref, 4.5, Z, s2w);
  double W = Z*sqrt(1-s2w);
  shift += TH(Z,W,s2w,h) - TH(Z,W,s2w,href);
  return shift;
}
